require 'smart_properties'

module Quilt
  module UIHelper
    class Component < SimpleDelegator
      # rubocop:disable Alias
      alias view_context __getobj__
      # rubocop:enable Alias

      include SmartProperties

      property :options, accepts: Hash, default: {}, required: true

      def initialize(view_context, **args)
        class_properties = self.class.properties.keys
        valid_args = args.slice(*class_properties)
        options = args.except(*class_properties)

        super(view_context, options: options, **valid_args)
      end

      def self.new(*args, **attrs, &block)
        super
      rescue SmartProperties::AssignmentError => error
        fallback_error = FallbackError.new(error)

        raise unless Quilt::UIHelper.render_fallback
        if Quilt::UIHelper.logger_callback.respond_to?(:call)
          Quilt::UIHelper.logger_callback.call(fallback_error)
        else
          Rails.logger.error(fallback_error)
        end

        fallback(*args, **attrs, &block)
      end

      def self.fallback(*)
        Fallback.new
      end

      def self.variants
        parent_variants = defined?(super) ? super : {}
        (parent_variants || {}).merge(@variants || {})
      end

      def self.variant(name, accepts: [true, false], **options)
        @variants ||= {}
        @variants[name] = Variant.new(name, Set.new(accepts))
        property name, accepts: accepts, **options
      end

      def self.base_css_class(class_name)
        define_method(:base_css_class) { class_name }
      end

      def base_css_class
        raise NotImplementedError, 'You need to define a base_css_class method for your component'
      end

      def variant_classes(*additional_classes)
        classes = [base_css_class]

        self.class.variants.each do |variant_key, variant|
          variant_value = self[variant_key]
          next unless variant_value
          classes << base_css_class_variant(variant_key, variant_value, variant)
        end

        [classes, additional_classes].flatten.compact
      end

      def ui_content_tag(name, *content_or_additional_options, **base_options, &block)
        additional_options = content_or_additional_options.extract_options!

        content = block_given? ? capture(&block) : safe_join(Array(content_or_additional_options))

        options = base_options.merge(additional_options) do |key, base_value, additional_value|
          case key
          when :class
            Array(base_value) + Array(additional_value)
          else
            additional_value
          end
        end

        content.blank? ? ''.html_safe : content_tag(name, options) { content }
      end

      protected

      def base_css_class_variant(variant_key, variant_value, variant)
        class_name = "#{base_css_class}--#{variant_key}"
        class_name << "-#{variant_value}" unless variant.boolean?
        bemerize(class_name)
      end

      def bemerize(css_class_name)
        css_class_name.gsub(/(?<!_)_(?!_)/, '-')
      end

      class Variant < Struct.new(:name, :values)
        BOOLEAN = Set.new([true, false]).freeze

        def boolean?
          values == BOOLEAN
        end
      end

      class Fallback
        def render(*)
          ''.html_safe
        end
      end

      class FallbackError < RuntimeError; end
    end
  end
end
